#lang typed/racket

(struct handle ([id : Number]) #:transparent)
(struct [T] heap ([data : (Immutable-HashTable handle T)]) #:transparent)
(struct [S T] eff ([state : S] [result : T]) #:transparent)

(define-type (eff-op State Result) (-> State (eff State Result)))

(: eff-bind
  (All [State Input Output]
    (->
      (eff-op State Input)
      (-> Input (eff-op State Output))
      (eff-op State Output)
    )
  )
)
(define (eff-bind o1 o2)
  (lambda ([h1 : State])
    (define h2+r (o1 h1))
    (define r (eff-result h2+r))
    (define h2 (eff-state h2+r))
    ((o2 r) h2)))


(: seq2
  (All [State Input Output]
    (->
      (eff-op State Input)
      (eff-op State Output)
      (eff-op State Output)
    )
  )
)
(define (seq2 op1 op2)
  (eff-bind op1 (lambda (x) op2)))

(: seq
  (All [State T]
    (->*
      ((eff-op State T))
      #:rest
      (eff-op State T)
      (eff-op State T)
    )
  )
)
(define (seq op . ops)
 (cond [(empty? ops) op]
        [else (seq2 op (apply seq ops))])
)


(: eff-pure
  (All [State T]
    (-> T (eff-op State T))
  )
)
(define (eff-pure x)
  (lambda (h) (eff h x)))

(: make-heap (All [T] (-> (heap T))))
(define (make-heap)
  (: h (Immutable-HashTable handle T))
  (define h (hash))
  (heap h))
(: heap-alloc (All [T] (-> (heap T) T (eff (heap T) handle))))
(define (heap-alloc h v)
  (define data (heap-data h))
  (define new-id (handle (hash-count data)))
  (define new-heap (heap (hash-set data new-id v)))
  (eff new-heap new-id))

(: heap-get (All [T] (-> (heap T) handle T)))
(define (heap-get h k)
  (hash-ref (heap-data h) k))

(: heap-put (All [T] (-> (heap T) handle T (heap T))))
(define (heap-put h k v)
  (define data (heap-data h))
  (cond
    [(hash-has-key? data k) (heap (hash-set data k v))]
    [else (error "Unknown handle!")]))
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(: prog1 (-> (heap Real) (eff (heap Real) handle)))
(define (prog1 h1)
  ;; `allocate x with 1`
  (define eff-x (heap-alloc h1 10))
  (define x (eff-result eff-x))
  (define h2 (eff-state eff-x))
  ;; `allocate y with 2`
  (define eff-y (heap-alloc h2 2))
  (define y (eff-result eff-y))
  (define h3 (eff-state eff-y))
  ;; `allocate z with (+ x y)`
  (heap-alloc h3 (+ (heap-get h3 y) (heap-get h3 x))))



(make-heap)



(: num
  (->
    ; The number
    Real
    ; The parameterize computation
    (eff-op (heap Real) handle)
    ;(-> (heap Real) (eff (heap Real) handle))
  )
)
(define (num n)
  (lambda (h)
    (heap-alloc h n)
  )
)

(: mult
  (->
    handle
    handle
    (eff-op (heap Real) handle)
;;     (-> (heap Real)
;;       (eff (heap Real) handle)
;;     )
  )
)
(define (mult x y)
  (lambda (h)
    (heap-alloc h (+ (heap-get h y) (heap-get h x)))
  )
)
(: prog2 (eff-op (heap Real) handle))
(define prog2
  ;; `allocate x with 1`
  ; (define* x (num 10))
  (eff-bind (num 10)
    (lambda ([x : handle])
      ;; `allocate y with 2`
      (eff-bind (num 2)
        (lambda ([y : handle])
          (mult x y)))))
)

(prog2 (make-heap))

(define-type Stack (Listof Real))

(: push (-> Real (eff-op Stack Void)))
(define (push n)
  (lambda ([s : Stack])
    (eff (cons n s) (void))
  )
)

(: pop (-> (eff-op Stack Real)))
(define (pop)
  (lambda ([s : Stack])
    (match s
      [(list h l ...)
        (eff l h)
      ]
    )
  )
)

(define-syntax do
  (syntax-rules (<-)
    ; Only one monadic-op, return it
    [(_ mexp) mexp]
    ; A binding operation
    [(_ var : ty <- mexp rest ...) (eff-bind mexp (lambda ([var : ty]) (do rest ...)))]
    ; No binding operator, just ignore the return value
    [(_ mexp rest ...)        (eff-bind mexp (lambda (_) (do rest ...)))]))


(: multiply (-> (eff-op Stack Void)))
(define (multiply)
  (do
    x : Real <- (pop)
    y : Real <- (pop)
    (push (* x y)))
)

(: prog5 (eff-op Stack Void))
(define prog5
  (do
    (push 2)
    (push 5)
    (multiply)
    (push 2)
    (multiply)))
(prog5 (list))
#|
;(: num (-> Real (-> (heap Real) (eff (heap Real) handle))))
(: num (-> Real (eff-op (heap Real) handle)))
(define (num x)
  (lambda ([h : (heap Real)])
    (heap-alloc h x))
)

;(: add (-> handle handle (-> (heap Real) (eff (heap Real) handle))))
(: add (-> handle handle (eff-op (heap Real) handle)))
(define (add x y)
  (lambda ([h : (heap Real)])
    (heap-alloc h
      (+ (heap-get h y) (heap-get h x)))))

(: prog2 (eff-op (heap Real) handle))
(define (prog2 h1)
  ;; `allocate x with 1`
  (define eff-x ((num 1) h1))
  (define x (eff-result eff-x))
  (define h2 (eff-state eff-x))
  ;; `allocate y with 2`
  (define eff-y ((num 2) h2))
  (define y (eff-result eff-y))
  (define h3 (eff-state eff-y))
  ;; `allocate z with (+ x y)`
  ((add x y) h3)
)

(: prog3 (eff-op (heap Real) handle))
(define prog3
  (eff-bind (num 1)
    (lambda ([x : handle])
      ;; allocate x with 2
      (eff-bind (num 2)
        (lambda ([y : handle])
          ;; allocate y with (+ x y)
          (add x y)))))
)

(: pop (-> (eff-op (Listof Real) Real)))
(define (pop)
  (lambda (stack )
    (eff (rest stack) (first stack)))
)

(: push (-> Real (eff-op (Listof Real) Void)))
(define (push n)
  (lambda (stack)
    (eff (cons n stack) (void))))

(: mult (-> (eff-op (Listof Real) Void)))
(define (mult)
  (eff-bind (pop)
    (lambda ([x : Real])
      (eff-bind (pop)
        (lambda ([y : Real])
          (push (* x y)))))))

(: prog4 (eff-op (Listof Real) Void))
(define prog4
  (eff-bind (push 2)
    (lambda (x1)
      (eff-bind (push 5)
        (lambda (x2)
          (eff-bind (mult)
            (lambda (x3)
              (eff-bind (push 2)
                (lambda (x4)
                  (mult))))))))))

(: prog4 (eff-op (Listof Real) Void))
(define prog5
  (seq (push 2)
       (push 5)
       (mult)
       (push 2)
       (mult)))
(define-syntax do
  (syntax-rules (<-)
    ; Only one monadic-op, return it
    [(_ mexp) mexp]
    ; A binding operation
    [(_ var : ty <- mexp rest ...) (eff-bind mexp (lambda ([var : ty]) (do rest ...)))]
    ; No binding operator, just ignore the return value
    [(_ mexp rest ...)        (eff-bind mexp (lambda (_) (do rest ...)))]))

(: mult2 (-> (eff-op (Listof Real) Void)))
(define (mult2)
  (do
    x : Real <- (pop)
    y : Real <- (pop)
    (push (* x y))))
|#
